#include "toText.h"
#include <iomanip>
#include <sstream>
#include "capd/capd/minmax.h"
#include "capd/interval/Interval.hpp"

FloatVector diameter(IntervalVector &v)
{
   FloatVector t(v.dimension());
   for(int i=0; i<v.dimension(); ++i)
      t[i] = v[i].rightBound() - v[i].leftBound();
   return t;
}

std::string initialConditions( int iter, const IntervalVector &x0, const IntervalVector &X,
                               const IntervalVector &x0_e, const IntervalVector &X_e)
{
   std::ostringstream str;
   str << std::setprecision(12)<<"\n==================================================\n"
       << " INITIAL CONDITIONS for iteration " << iter <<" \n==================================================\n"
       << "\n In reduced space \n\n x0 = " << x0
       << "\n\n X = " <<X <<"\n\n diam X "<<(size(X).rightBound())
       << "\n\n In full space\n\n  E(x0) = " <<x0_e
       << "\n\n E(X) = " <<X_e<<"\n\n diam E(X) "<<(size(X_e).rightBound())<<std::endl;

   return str.str();
}

std::string saveInitialConditions( int iter, IntervalVector &x0, IntervalVector &X)
{
   std::ostringstream str;

   str << std::setprecision(12) << "\n==================================================\n"
       << " INITIAL CONDITIONS for iteration " << iter <<" \n==================================================\n"
       << "\n\n x0 = " << x0
       << "\n\n X = " <<X <<"\n\n diam X "<< size(X).rightBound();

   return str.str();

}
// For intervalMatrix A returns
//  max - absolute value of the biggest entry
//   average - average  of  entries value
//   maxDiam - maximal diameter of its coordinates
//  averageDiam - average diameter

void  matrixTest(IntervalMatrix &A, ScalarType & max, ScalarType &average, ScalarType &maxDiam, ScalarType &averageDiam)
{
   max = capd::abs(A[0][0]).rightBound();
   average =0;
   averageDiam =0;
   maxDiam = 0;

   for(int i = 0; i < A.numberOfRows(); i++)
   {
       for(int j = 0; j < A.numberOfColumns(); j++)
       {
          if(max < capd::abs(A[i][j]).rightBound())
             max = capd::abs(A[i][j]).rightBound();
          average+= capd::abs(A[i][j]).rightBound();

          ScalarType diam = (A[i][j]).rightBound() - (A[i][j]).leftBound() ;
          if(maxDiam < diam)
             maxDiam = diam;
          averageDiam += diam;

       }
   }
   averageDiam /= (A.numberOfRows()*A.numberOfColumns());
   average /= (A.numberOfRows()*A.numberOfColumns());

}


// It saves to file fileName information about Matrix
//  such as maximal and average value and diameter of its cooeficient
//
std::string toText(const char* matrixName, IntervalMatrix &A, int prec, bool show)
{

  ScalarType max, average, maxDiam, averageDiam;
  matrixTest(A,  max, average, maxDiam, averageDiam);
  std::ostringstream str;
  if(prec)
        str << std::setprecision(prec);

  str << "\n matrix " << matrixName;
  if(show) str << "\n" << A;
  str << "\n max value        =" << max
      << "\n average value    = " << average
      << "\n max diameter     = " << maxDiam
      << "\n average diameter = " << averageDiam;
  IntervalVector d = A[0];
  str << " \n diameters of first row of " << matrixName << "\n  "<< diameter(d)<< std::endl;
  return str.str();
}

std::string toText(const char* vectorName, IntervalVector &v, int prec, bool show)
{

   std::ostringstream str;
   if(prec)
      str << std::setprecision(prec);
   str <<"\n vector " << vectorName
       <<"\n"  << v
       << "\n max diameter     = " << size(v).rightBound()
       << " \n diam :" << diameter(v) << std::endl;
   return str.str();
}

std::string toText(const char* vectorName, FloatVector &v, int prec, bool show)
{

  std::ostringstream str;
  if(prec)
    str << std::setprecision(prec);
  str <<"\n " << vectorName
      <<"\n"  << v;
  return str.str();
}

std::string toText(const char* intervalName, interval &v, int prec, bool show)
{
   std::ostringstream str;
   if(prec)
      str << std::setprecision(prec);
   str  << intervalName
        <<"\n" << v
        << "\n diameter     = " << diam(v);
   return str.str();
}


std::string parametersToText(const char* text,   ScalarType step, int order)
{
   std::ostringstream str;
   str << std::endl <<  text << std::endl
       << "  step = " << step <<"    order = " << order << std::endl;
   return str.str();
}

std::string parametersToText(const char* text,    ScalarType step, int order, ScalarType X_step, int X_order)
{
   std::ostringstream str;
   str << std::endl <<  text << std::endl << "for a point: step = " << step
       <<"    order = " << order << std::endl;
   str << "for a set: step = " << X_step << "    order = " << X_order << std::endl;
   return str.str();
}
