// This is TIMEUSED.CPP. Copyright (C) 2002-2004 by Pawel Pilarczyk.
// It contains an implementation of the functions declared in "timeused.h".
// This is free software. No warranty. Consult 'license.txt' for details.
// Started on March 23, 2002. Last revision: June 3, 2004.


// #define _TIMEUSED_CPP_

#include "options.h"
#include "timeused.h"

#include <ctime>
#include <iostream>
#include <iomanip>

#ifdef ppUNIX
#include <sys/time.h>
#if ((!defined (ppDOS)) || defined (DJGPP))
#include <sys/resource.h>
#endif
#include <unistd.h>
#endif


// --------------------------------------------------
// ------------------- TIMEUSED ---------------------
// --------------------------------------------------

inline double currentcpu ()
// Return the current time in seconds.
{
	#if ((!defined (ppDOS)) || defined (DJGPP))
	struct rusage usage;
	getrusage (RUSAGE_SELF, &usage);
	return usage. ru_utime. tv_sec + usage. ru_utime. tv_usec / 1e6 +
		usage. ru_stime. tv_sec + usage. ru_stime. tv_usec / 1e6;
	#else
	return (double) clock () / CLK_TCK;
	#endif
} /* currenttime */

inline std::ostream &showseconds (std::ostream &out, double seconds, int precision)
{
	int prec = out. precision ();
	out. setf (std::ios::fixed);

	out << std::setprecision (precision) << seconds << " sec";
//	if (seconds < 60);
	if (seconds < 3600)
		out << " (" << std::setprecision (precision + 1) <<
			(seconds / 60) << " min)";
	else
		out << " (" << std::setprecision (precision + 1) <<
			(seconds / 3600) << " hours)";

	out. unsetf (std::ios::fixed);
	out. precision (prec);

	return out;
} /* showseconds */

std::ostream &operator << (std::ostream &out, const timeused &t)
{
	double c_sec = currentcpu () - t. cpu0;
	time_t t1;
	time (&t1);
	double t_sec = difftime (t1, t. t0);

	int precision = 0;
	if (c_sec < 100)
		precision ++;
	if (c_sec < 10)
		precision ++;

	showseconds (out, c_sec, precision);

	if (t_sec > c_sec + 1)
	{
		out << " out of ";
		showseconds (out, t_sec, 0);
	}

	return out;
} /* operator << */

timeused &timeused::reset ()
{
	cpu0 = currentcpu ();
	time (&t0);
	return *this;
} /* timeused::reset */

timeused::operator double ()
{
	if (cpu0 >= 0)
		return (currentcpu () - cpu0);
	time_t t1;
	time (&t1);
	return difftime (t1, t0);
} /* timeused::operator double */

// --------------------------------------------------

timeused program_time (std::cout);

